<?php
/**************************************************************
 * Subset of Sass for PHP
 * © 2009, artoodetoo.
 *
 * Sass homepage in http://sass-lang.com/
 **************************************************************/

error_reporting(E_ALL);

class SassFilter
{
	const TAB = 4;
	// Charset for lexer
	const ID_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_';
	const VAR_PATTERN = '\![a-zA-Z0-9\-_]+';
	// Sass string kind
	const VAR_DEF	= 1;
	const MIXIN_DEF	= 2;
	const MIXIN_USE	= 3;
	const RULE_DEF	= 4;
	const ATTR_PREF	= 5;
	const ATTR_DEF	= 6;
	const ATTR_CALC	= 7;
	const COMMAND	= 8;
	// Output style
	const O_EXPANDED	= 0;
	const O_NESTED		= 1;
	const O_COMPACT		= 2;
	const O_COMPRESSED	= 3;

	private static $_levels;
	private static $_names;
	private static $_variables;
	private static $_mixins;
	private static $_rules;

	private static function _cleanStr($s)
	{
		if(($p = strpos($s, '//')) !== FALSE)
			$s = substr($s, 0, $p);
		$s = rtrim($s);
		while (($p = strpos($s, "\t")) !== FALSE)
			$s = substr_replace($s, str_repeat(' ', self::TAB - $p % self::TAB), $p, 1);
		return $s;
	}

	private static function &_valueByName($name, $kind = self::VAR_DEF)
	{
		$ret = NULL;
		foreach (self::$_names as $key => $item)
			if ($item[0] == $name && $item[1] == $kind)
			{
				switch ($kind) {
				case self::VAR_DEF:		$ret =& self::$_variables[$key]; break 2;
				case self::MIXIN_DEF:	$ret =& self::$_mixins[$key]; break 2;
				case self::RULE_DEF:	$ret =& self::$_rules[$key]; break 2;
				}
			}
		return $ret;
	}

	private static function &_blockByOffset($offs)
	{
		$ret = NULL;
		foreach (self::$_levels as $o => $item)
		{
			if ($o >= $offs)
				break;
			if ($item[1] == self::MIXIN_DEF)
				$ret =& self::$_mixins[$item[0]];
			else if ($item[1] == self::RULE_DEF)
				$ret =& self::$_rules[$item[0]];
		}
		return $ret;
	}

	private static function _attrNameByOffset($offs)
	{
		$ret = '';
		foreach (self::$_levels as $o => $item)
		{
			if ($o > $offs) break;
			if (self::ATTR_PREF <= $item[1] && $item[1] <= self::ATTR_CALC)
				$ret = ($ret != '' ? $ret.'-' : '') . self::$_names[$item[0]][0];
		}
		return $ret;
	}

	private static function _ruleNameByOffset($offs)
	{
		$childSelector = self::$_names[self::$_levels[$offs][0]][0];
		$parentSelector = NULL;

		foreach (self::$_levels as $o => $item)
		{
			if ($o >= $offs) break;
			if ($item[1] == self::RULE_DEF)
				$parentSelector = self::$_names[$item[0]][0];
		}
		if (!is_null($parentSelector))
		{
			$parentArray = explode(',', $parentSelector);
			$childArray = explode(',', $childSelector);
			foreach ($parentArray as $parent)
				foreach ($childArray as &$child)
					if (($ampPos = strpos($child, '&')) !== FALSE)
						$child = substr_replace($child, $parent, $ampPos, 1);
					else
						$child = $parent . ' ' . $child;
			$childSelector = str_replace('  ', ' ', implode(', ', $childArray));
		}

		return $childSelector;
	}

	private static function _getVar($matches)
	{
		return self::_valueByName(substr($matches[0], 1), self::VAR_DEF);
	}

	// TODO: calculate expression
	private static function _expression($s)
	{
		return preg_replace_callback('/'.self::VAR_PATTERN.'/', array('self', '_getVar'), $s);
	}

	private static function _parseStr($s)
	{
		$value = NULL;
		$s = ltrim($s);
		switch ($s{0}) {
		case '@':
			$name = substr($s, 1, $l = strspn($s, self::ID_CHARS, 1));
			$value = ltrim(substr($s, 1 + $l));
			$kind = self::COMMAND;
			break;
		case '!':
			$name = substr($s, 1, strspn($s, self::ID_CHARS, 1));
			$value = ltrim(substr($s, 1 + strpos($s, '=')));
			$kind = self::VAR_DEF;
			break;
		case '=':
			$name = substr($s, 1, strspn($s, self::ID_CHARS, 1));
			$kind = self::MIXIN_DEF;
			break;
		case '+':
			$name = substr($s, 1, strspn($s, self::ID_CHARS, 1));
			$value = self::_valueByName($name, self::MIXIN_DEF);
			$kind = self::MIXIN_USE;
			break;
		case ':':
			$name = substr($s, 1, $l = strspn($s, self::ID_CHARS, 1));
			$value = ltrim(substr($s, 1 + $l));
			if ($value == '')
				$kind = self::ATTR_PREF;
			else if ($value{0} == '=')
			{
				$value = ltrim(substr($value, 1));
				$kind = self::ATTR_CALC;
			}
			else
				$kind = self::ATTR_DEF;
			break;
		default:
			$name = $s;
			$kind = self::RULE_DEF;
		}
		return array($kind, $name, $value);
	}

	private static function _processStr(&$s, $key)
	{
		$offs = strspn($s, ' ');
		list($kind, $name, $value) = self::_parseStr($s);
		self::$_levels[$offs] = array($key, $kind);
		self::$_names[$key] = array($name, $kind, $offs);

		switch ($kind) {
		case self::VAR_DEF:
			self::$_variables[$key] = self::_expression($value);
			break;
		case self::MIXIN_DEF:
			self::$_mixins[$key] = array();
			break;
		case self::MIXIN_USE:
			$target =& self::_blockByOffset($offs);
			foreach ($value as $line)
				$target[] = $line;
			break;
		case self::RULE_DEF:
			self::$_names[$key][0] = self::_ruleNameByOffset($offs);
			self::$_rules[$key] = array();
			break;
		case self::ATTR_PREF:
			break;
		case self::ATTR_DEF:
			$target =& self::_blockByOffset($offs);
			$finalName = self::_attrNameByOffset($offs);
			$target[] = array($finalName, $value);
			break;
		case self::ATTR_CALC:
			$target =& self::_blockByOffset($offs);
			$finalName = self::_attrNameByOffset($offs);
			$target[] = array($finalName, self::_expression($value));
			break;
		default:
			// TODO: @command processing
		}
/*		// debug
		_debugPrint($offs, $key, $kind, $name, $s); */
	}

	private static function _processAll($sass)
	{
		$_levels = $_names = $_variables = $_mixins = $_rules = array();

		$lines = array_filter(array_map(array('self', '_cleanStr'), explode("\n", $sass)), 'strlen');
		array_walk($lines, array('self', '_processStr'));
	}

	public static function pipe($sass, $style= O_NESTED)
	{
		self::_processAll($sass);

		$ret = '';
		switch ($style) {
		case self::O_EXPANDED:
			foreach (self::$_rules as $key => $attrs)
			{
				if (empty($attrs)) continue;
				$ret .= self::$_names[$key][0] . ' { ';
				foreach ($attrs as $attr)
					$ret .= "\n\t" . $attr[0] . ': ' . $attr[1] . ';';
				$ret .= "\n}\n\n";
			}
			break;
		case self::O_NESTED:
			foreach (self::$_rules as $key => $attrs)
			{
				$tab = str_repeat(' ', self::$_names[$key][2]);
				$ret .= $tab . self::$_names[$key][0] . ' { ';
				foreach ($attrs as $attr)
					$ret .= "\n" . $tab . '  ' . $attr[0] . ': ' . $attr[1] . ';';
				$ret .= "}\n";
			}
			break;
		case self::O_COMPACT:
			foreach (self::$_rules as $key => $attrs)
			{
				if (empty($attrs)) continue;
				$ret .= self::$_names[$key][0] . ' { ';
				foreach ($attrs as $attr)
					$ret .= "" . $attr[0] . ': ' . $attr[1] . '; ';
				$ret .= "}\n";
			}
			break;
		case self::O_COMPRESSED:
			foreach (self::$_rules as $key => $attrs)
			{
				if (empty($attrs)) continue;
				$ret .= self::$_names[$key][0] . '{';
				foreach ($attrs as $attr)
					$ret .= "" . $attr[0] . ':' . $attr[1] . ';';
				$ret .= "}";
			}
			break;
		}
		return $ret;
	}
}


/*
// debug
function _debugPrint($offs, $key, $kind, $name, $s)
{
	switch ($kind) {
	case SassFilter::VAR_DEF:	$k = 'VAR_DEF'; break;
	case SassFilter::MIXIN_DEF:	$k = 'MIXIN_DEF'; break;
	case SassFilter::MIXIN_USE:	$k = 'MIXIN_USE'; break;
	case SassFilter::RULE_DEF:	$k = 'RULE_DEF'; break;
	case SassFilter::ATTR_DEF:	$k = 'ATTR_DEF'; break;
	case SassFilter::ATTR_PREF:	$k = 'ATTR_PREF'; break;
	case SassFilter::ATTR_CALC:	$k = 'ATTR_CALC'; break;
	}
	echo sprintf("%2d: (%3d, %-9s, %-15s) %s\n", $offs, $key, $k, $name, $s);
}
*/

$sass = file_get_contents('./sass/test2.sass');

echo "<pre>\n";
echo SassFilter::pipe($sass, SassFilter::O_EXPANDED);
echo "\n</pre>\n";

?>